"""Fixtures and utilities for integration tests."""

# better traceback formatting with rich, if available
import re
from collections.abc import Generator
from collections.abc import Sequence
from pathlib import Path
from re import Pattern

import dotenv
import pytest
from loguru import logger as log
from responses import RequestsMock
from spectrumx.client import Client
from spectrumx.gateway import API_PATH
from spectrumx.gateway import API_TARGET_VERSION
from spectrumx.gateway import Endpoints

# matches UUIDv4 strings for passthru validation
uuid_v4_regex = (
    r"[0-9a-fA-F]{{8}}-[0-9a-fA-F]{{4}}-[0-9a-fA-F]"
    r"{{4}}-[0-9a-fA-F]{{4}}-[0-9a-fA-F]{{12}}"
)

dir_integration_data = Path(__file__).parent / "data"
assert dir_integration_data.exists(), (
    f"Integration test data directory '{dir_integration_data}' does not exist."
)

# add more hostnames here to bypass `responses` when running integration tests.
# these servers will receive the actual requests from integration tests:
passthru_hostnames = [
    "http://localhost:80",
    "https://sds-dev.crc.nd.edu:443",
    # "https://sds-staging.example.com:443",    # add your instance here  # noqa: ERA001
    #
    # To change the host used in these tests, see integration.env.example in this dir.
    #
    # Note: running integration tests against sds.crc.nd.edu is not recommended;
    #   but you may redirect requests to this name to a different machine by
    #   changing your local /etc/hosts first. This is useful if you're also
    #   testing your Traefik config, for example.
]


class PassthruEndpoints:
    """Wrapper to return commonly used passthru hostnames.

    Passthru hostnames are a collection of complete endpoints -- protocol + (sub+)domain
        + port + path -- that are used to bypass the `responses` package in integration
        tests. All endpoints matched by these URLs will be allowed to pass
        through to the real service behind, so that the integration can be tested.
    The method (GET, POST, etc) is not part of a passthru, so some goals share the same
        passthru URL, such as the file metadata download (GET) and file uploads (POST).
    Any requests not in the passthru list will be intercepted by `responses`, and,
        unless they are mocked, that will raise an error.

    Usage:

    Decorate a test function with the following:

    ```
    @pytest.mark.usefixtures("_without_responses")
    @pytest.mark.parametrize(
        "_without_responses",
        argvalues=[
            PassthruEndpoints.authentication(),
            # other passthru endpoints here
        ],
        # tell pytest to pass the parameters to the fixture, \
        #   instead of the test function directly:
        indirect=True,
    )
    ```

    """

    @staticmethod
    def authentication() -> list[str]:
        """Passthrough for authentication."""
        return [
            f"{hostname_and_port}{API_PATH}{API_TARGET_VERSION}{Endpoints.AUTH}"
            for hostname_and_port in passthru_hostnames
        ]

    @staticmethod
    def file_content_checks() -> list[str]:
        """Passthrough for file content checks."""
        return [
            f"{hostname_and_port}{API_PATH}{API_TARGET_VERSION}{Endpoints.FILE_CONTENTS_CHECK}"
            for hostname_and_port in passthru_hostnames
        ]

    @classmethod
    def file_uploads(cls) -> list[str]:
        """Passthrough for upload of file metadata + contents.

        Alias for file_meta_download_or_upload() to make usage more explicit.
        """
        return cls.file_meta_download_or_upload()

    @staticmethod
    def file_meta_download_or_upload() -> list[str]:
        """Passthrough for file metadata download or file uploads."""
        return [
            f"{hostname_and_port}{API_PATH}{API_TARGET_VERSION}{Endpoints.FILES}"
            for hostname_and_port in passthru_hostnames
        ]

    @staticmethod
    def file_content_download() -> list[Pattern[str]]:
        """Passthrough for file content download."""
        return [  # file content download
            re.compile(
                rf"{hostname_and_port}{API_PATH}{API_TARGET_VERSION}{Endpoints.FILES}/{uuid_v4_regex}/download"
            )
            for hostname_and_port in passthru_hostnames
        ]

    @staticmethod
    def capture_reading() -> list[Pattern[str]]:
        """Passthrough for file content download."""
        return [  # file content download
            re.compile(
                rf"{hostname_and_port}{API_PATH}{API_TARGET_VERSION}{Endpoints.CAPTURES}/{uuid_v4_regex}/"
            )
            for hostname_and_port in passthru_hostnames
        ]

    @staticmethod
    def capture_creation() -> list[str]:
        """Passthrough for capture creation."""
        return [
            f"{hostname_and_port}{API_PATH}{API_TARGET_VERSION}{Endpoints.CAPTURES}"
            for hostname_and_port in passthru_hostnames
        ]

    @staticmethod
    def capture_deletion() -> list[Pattern[str]]:
        """Passthrough for capture deletion by UUID."""
        return [
            re.compile(
                rf"{hostname_and_port}{API_PATH}{API_TARGET_VERSION}{Endpoints.CAPTURES}/{uuid_v4_regex}/"
            )
            for hostname_and_port in passthru_hostnames
        ]

    @classmethod
    def capture_listing(cls) -> list[str]:
        """Passthrough for capture listing (same as capture creation)."""
        return cls.capture_creation()

    @classmethod
    def capture_search(cls) -> list[str]:
        """Passthrough for capture search (same as capture creation)."""
        return cls.capture_creation()

    @staticmethod
    def file_deletion() -> list[Pattern[str]]:
        """Passthrough for file deletion by UUID."""
        return [
            re.compile(
                rf"{hostname_and_port}{API_PATH}{API_TARGET_VERSION}{Endpoints.FILES}/{uuid_v4_regex}/"
            )
            for hostname_and_port in passthru_hostnames
        ]

    @staticmethod
    def all_passthru() -> list[str | Pattern[str]]:
        """Returns all passthru endpoints for debugging."""
        return (
            PassthruEndpoints.authentication()
            + PassthruEndpoints.file_content_checks()
            + PassthruEndpoints.file_uploads()
            + PassthruEndpoints.file_meta_download_or_upload()
            + PassthruEndpoints.file_content_download()
            + PassthruEndpoints.capture_creation()
            + PassthruEndpoints.file_deletion()
        )


@pytest.fixture
def _integration_setup_teardown() -> Generator[None]:
    """Fixture to set up and tear down integration tests."""
    # setup code for integration tests

    # yield control to the test
    yield

    # teardown code for integration tests

    return


@pytest.fixture
def _without_responses(
    request: pytest.FixtureRequest, responses: RequestsMock
) -> Generator[None]:
    """Adds URL bypasses to the responses package.

    Used in integration tests like:

    ```
    @pytest.mark.integration    # optional
    @pytest.mark.usefixtures("_integration_setup_teardown")   # optional
    @pytest.mark.parametrize(
        "_without_responses",
        [
            [   # list of URLs to allow bypass (the real service is called)
                "https://sds-dev.crc.nd.edu:443{API_PATH}v1/auth",
                "http://localhost:80{API_PATH}v1/auth",
            ]
        ],
        # tell pytest to pass the parameters to this
        #   fixture, instead of the test function directly:
        indirect=True,
    )
    ```
    """

    # setup
    responses.reset()
    bypassed_urls: Sequence[str | Pattern[str]] = request.param
    log.debug("Test allowing request bypass of the URLs:")
    for url in bypassed_urls:
        log.debug(f"  - {url}")
        responses.add_passthru(url)

    # yield control to the test
    yield

    # teardown
    responses.reset()


@pytest.fixture
def integration_client() -> Client:
    """Fixture to create a Client instance for integration testing."""
    integration_config = {
        "DRY_RUN": False,  # disable dry run for integration tests
        "HTTP_TIMEOUT": 30,
        "SDS_HOST": "sds-dev.crc.nd.edu",
    }
    _integration_client = Client(
        host="sds-dev.crc.nd.edu",
        env_config=integration_config,
        env_file=Path("tests")
        / "integration"
        / "integration.env",  # contains the SDS_SECRET_TOKEN
        verbose=True,
    )
    assert (
        _integration_client._config.api_key is not None  # noqa: SLF001 # pyright: ignore[reportPrivateUsage]
    ), "Client didn't load the API key."
    assert _integration_client.dry_run is False, "Dry run mode should be disabled."
    return _integration_client


@pytest.fixture
def inline_auth_integration_client() -> Client:
    """Client instance for integration testing authenticated with inline."""
    env_file = Path("tests") / "integration" / "integration.env"
    env_file_loaded = dotenv.dotenv_values(env_file, verbose=True)
    assert "SDS_SECRET_TOKEN" in env_file_loaded, (
        f"No SDS_SECRET_TOKEN found in '{env_file}'."
    )
    sds_token = env_file_loaded["SDS_SECRET_TOKEN"]
    _integration_client = Client(
        host="sds-dev.crc.nd.edu",
        env_config={
            "SDS_SECRET_TOKEN": sds_token,
            "DRY_RUN": False,
        },
        env_file=None,
        verbose=True,
    )
    assert (
        _integration_client._config.api_key is not None  # noqa: SLF001 # pyright: ignore[reportPrivateUsage]
    ), "Client didn't load the API key."
    assert _integration_client.dry_run is False, "Dry run mode should be disabled."
    return _integration_client


@pytest.fixture
def rh_sample_top_level_dir() -> Path:
    """Fixture to provide the RadioHound sample top-level directory."""
    return dir_integration_data / "captures" / "radiohound"


@pytest.fixture
def drf_sample_top_level_dir() -> Path:
    """Fixture to provide the Digital-RF sample top-level directory."""
    return dir_integration_data / "captures" / "drf" / "westford-vpol"



"""Tests versioning of SpectrumX."""

import re


def test_version_exists() -> None:
    """Tests the version of the SpectrumX SDK exists."""
    from spectrumx import __version__

    assert isinstance(__version__, str), "Version number must be a string"


def test_version_sem_ver() -> None:
    """Tests the version number follows semantic versioning.

    Following https://peps.python.org/pep-0440/#public-version-identifiers
    """
    from spectrumx import __version__

    sem_ver_pattern: re.Pattern[str] = re.compile(
        r"^\d+(\.\d+)*([abrc]\d+)?(\.post\d+)?(\.dev\d+)?$"
    )
    assert sem_ver_pattern.match(__version__), (
        "Version number must follow semantic versioning with optional "
        "pre-release, post-release, or dev-release segments"
    )


def test_version_pyproject() -> None:
    """Tests the version is the same as in pyproject.toml."""
    from importlib.metadata import version

    from spectrumx import __version__

    assert __version__ == version(
        "spectrumx",
    ), "Version number must match pyproject.toml"



"""Custom exceptions for the SDS SDK."""

import sys
from collections.abc import Sequence
from typing import Any
from typing import Generic
from typing import TypeVar

from spectrumx.utils import log_user_error

# python 3.10 backport
if sys.version_info < (3, 11):  # noqa: UP036
    from typing_extensions import Self  # noqa: UP035 # Required backport
else:
    from typing import Self

from loguru import logger as log

log.trace("Placeholder log avoid reimporting or resolving unused import warnings.")


class SDSError(Exception):
    """Base class for all SDS errors."""

    def __init__(self, message) -> None:
        self.message = message

    def __str__(self) -> Any:
        return self.message


class AuthError(SDSError):
    """Issue with user authentication against SDS."""


class NetworkError(SDSError):
    """Issue with the client connection to the SDS service."""


class ServiceError(SDSError):
    """Issue with the SDS service."""


class FileError(SDSError):
    """Issue interacting with a file in SDS."""


class CaptureError(SDSError):
    """Issue interacting with a capture in SDS."""


class DatasetError(SDSError):
    """Issue interacting with a dataset in SDS."""


class ExperimentError(SDSError):
    """Issue interacting with an experiment in SDS."""


T = TypeVar("T")
UnsetT = TypeVar("UnsetT", bound="Unset")


class Unset:
    """A placeholder for unset values and allow None to be a valid value."""

    _instance: None | Self = None

    def __new__(cls, *args, **kwargs) -> "Unset":
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __repr__(self) -> str:
        return "<unset>"

    def __bool__(self) -> bool:
        return False

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Unset)


_unset = Unset()
_unset_alt = Unset()
assert _unset is _unset_alt, "Singleton self-test failed"
assert not _unset, "Unset should be falsy"
del _unset_alt


class Result(Generic[T]):
    """Either packs a value (success) or an exception (failure).

    Useful when running several operations that might
        fail, but you'd like to handle failures later.

    Notice the _constructor_ will raise if any are true:
        1. Both `value` and `exception` are provided.
        2. Neither `value` nor `exception` are provided.
        3. `value` is an instance of `Exception`.
        4. `exception` is not an instance of `Exception`.
        5. `error_info` is provided without an `exception`.
           The opposite is allowed.

    Also, `None` is a valid `value` and will be considered a success.

    ### READING a `Result`

    There are several ways to use it. Assuming `result` is an instance of this class:

    1. By calling: `result()` either returns the value or re-raises the exception.
       You can also call `result.unwrap()` for the same effect.
    2. By checking truthfulness: `bool(result) is True` means success.
       This can be used to filter them:
       `failed_jobs = [result for result in all_jobs if not result]`.
    3. To access the value: `my_val = result.value_or(default_value)`.
    4. Or the exception: `my_exc = result.exception_or(default_exception)`.
    6. When failed, see `result.error_info` to get more information if provided.

    ### CREATING a `Result`

    ```
    def dangerous_fn(asset_id: str) -> int:
        if random.random() < 0.1:
            raise KnownError("Something went wrong")
        return hash(asset_id)

    def wrap_dangerous_fn(asset_id) -> Result[int]:
        try:
            return Result(value=dangerous_fn(asset_id))
        except KnownError as e:
            error_info = {
                "reason": "It happens ...",
                "asset_id": asset_id,
            }
            return Result(exception=e, error_info=error_info)
    ```

    #### These examples will raise

    ```
    Result()                                    # empty result
    Result(value=123, exception=RuntimeError()) # both value and exception
    Result(value=123, error_info={})            # error_info without exception
    Result(exception=123)                       # exception is not an Exception
    Result(value=RuntimeError())                # value is an Exception
    ```

    #### These are valid

    ```
    Result(value=123)                               # success
    Result(value=None)                              # success with None
    Result(exception=RuntimeError())                # simple failure
    Result(exception=RuntimeError(), error_info={}) # failure with extra info

    Result(exception=RuntimeError(), error_info=True)
    # allowed but not recommended (`error_info` is not a dict)
    ```

    > May your results be successful!

    """

    def __init__(
        self,
        *,
        value: T | Unset = _unset,
        exception: Exception | Unset = _unset,
        error_info: dict[str, Any] | None = None,
    ) -> None:
        """Initializes a Result object.

        Args:
            value: The value to pack.
            exception: The exception to pack.
            error_info: Additional information about the error.
        """
        if value is _unset and exception is _unset:  # pragma: no cover
            msg = "Either value or exception must be provided."
            raise ValueError(msg)
        if value is not _unset and exception is not _unset:  # pragma: no cover
            msg = "Only one of value or exception can be provided."
            raise ValueError(msg)
        if exception is not _unset and not isinstance(
            exception, Exception
        ):  # pragma: no cover
            msg = "Exception must be an instance of Exception."
            raise ValueError(msg)
        if exception is _unset and error_info is not None:  # pragma: no cover
            msg = "Error info can only be provided with an exception."
            raise ValueError(msg)
        if value is not _unset and isinstance(value, Exception):
            msg = "Value cannot be an instance of Exception."
            raise ValueError(msg)
        # While we encourage `error_info` being a dict[str, Any], we're deliberately
        # not validating its type here to avoid risking the failure of longer-running
        # scripts for a minor issue. If the reader of `error_info` accepts a different
        # type, they have a chance to handle it.
        self._value: T = value
        self._exception: Exception = exception
        self.error_info: dict[str, Any] = error_info if error_info is not None else {}

    def __bool__(self) -> bool:
        _value_is_set = self._value is not _unset
        _exc_is_set = self._exception is not _unset
        assert _value_is_set != _exc_is_set, "Invalid Result state"
        return _value_is_set and not _exc_is_set

    def __str__(self) -> str:
        return (
            f"Result(value={self._value})"
            if self
            else f"Result(exception={self._exception})"
        )

    def __repr__(self) -> str:
        return self.__str__()

    def __call__(self) -> T:
        """Returns the value or raises the exception."""
        if self:
            return self._value
        raise self._exception

    def value_or(self, default: T) -> T:
        """Returns the wrapped value or a default value when result is an exception."""
        return self._value if self else default

    def exception_or(self, default: Exception | None) -> Exception | None:
        """Returns the wrapped exception or a default one when result is a value."""
        return self._exception if not self else default

    def unwrap(self) -> T:
        """Alias for `self()`. Raises if the result is an exception."""
        return self()


def process_upload_results(
    upload_results: Sequence[Result],  # pyright: ignore[reportMissingTypeArgument]
    *,
    verbose: bool,
    raise_on_error: bool,
) -> bool:
    """Process the results of a batch upload, raising if any has failed.

    Args:
        upload_results:     A sequence of Result objects.
        verbose:            Whether to log errors.
        raise_on_error:     Whether to raise an exception on failure.
    Raises:
        The first SDSError wrapped as upload_results if any upload has failed and
        raise_on_error is True.
    Returns:
        bool: True if all uploads were successful, False otherwise.
    """
    successful_uploads = [res.unwrap() for res in upload_results if res]
    failed_uploads = [res for res in upload_results if not res]
    if failed_uploads:
        if raise_on_error:
            # then raise the first exception in failed results
            for res in failed_uploads:
                exception = res.exception_or(SDSError("File Upload Failed"))
                if exception:
                    raise exception
        if verbose:
            log_user_error(f"Failed to upload the following files: {failed_uploads}")
        return False
    if not successful_uploads:
        if verbose:
            log_user_error("No files uploaded, unable to create capture.")
        return False
    return True



"""Tests for the File model."""

# pylint: disable=redefined-outer-name

from datetime import datetime
from datetime import timedelta
from pathlib import PurePosixPath
from typing import Any

import pytest
from pydantic import BaseModel
from pytz import UTC
from pytz import timezone
from spectrumx.models.files import File
from spectrumx.models.files import PermissionRepresentation
from spectrumx.models.files import UnixPermissionStr


@pytest.fixture
def file_properties() -> dict[str, Any]:
    """Fixture to create a dictionary of file properties."""
    tz = timezone("UTC")
    return {
        "name": "test_file",
        "directory": PurePosixPath("/my/files/are/here"),
        "media_type": "text/plain",
        "permissions": "-w-rw-r--",
        "size": 111_222_333,
        "created_at": datetime.now(tz=tz) - timedelta(days=14),
        "updated_at": datetime.now(tz=tz),
        "expiration_date": datetime.now(tz=UTC) + timedelta(days=366),
    }


def test_file_created(file_properties: dict[str, Any]) -> None:
    """Test that a file can be created correctly."""
    new_file = File(
        **file_properties,
    )
    for key, value in file_properties.items():
        assert getattr(new_file, key) == value, (
            f"{key} does not match: {getattr(new_file, key)} != {value}"
        )


def test_file_path(file_properties: dict[str, Any]) -> None:
    """Test that the path property returns the correct path."""
    new_file = File(
        **file_properties,
    )
    assert new_file.path == file_properties["directory"] / file_properties["name"]


def test_chmod_props(file_properties: dict[str, Any]) -> None:
    """Test that the chmod_props property returns the correct chmod properties."""
    new_file = File(
        **file_properties,
    )
    assert new_file.chmod_props == "0o264"  # 264 (octal) <=> '-w-rw-r--'


# Permission Struct
def test_permission_deserialization() -> None:
    class Model(BaseModel):
        permission: UnixPermissionStr

    a = Model(permission="rwx------")
    assert a.permission == "rwx------"

    b = Model(permission=0o755)
    assert b.permission == "rwxr-xr-x"


def test_permission_serialization() -> None:
    class Model(BaseModel):
        permission: UnixPermissionStr

    a = Model(permission="rwx------")
    assert a.model_dump(context={"mode": "string"}) == {"permission": "rwx------"}

    assert a.model_dump(context={"mode": PermissionRepresentation.STRING}) == {
        "permission": "rwx------"
    }

    assert a.model_dump(context={"mode": "octal"}) == {"permission": "0o700"}

    assert a.model_dump(context={"mode": PermissionRepresentation.OCTAL}) == {
        "permission": "0o700"
    }

    b = Model(permission=0o755)
    assert b.model_dump(context={"mode": "string"}) == {"permission": "rwxr-xr-x"}

    assert b.model_dump(context={"mode": PermissionRepresentation.STRING}) == {
        "permission": "rwxr-xr-x"
    }

    assert b.model_dump(context={"mode": "octal"}) == {"permission": "0o755"}

    assert b.model_dump(context={"mode": PermissionRepresentation.OCTAL}) == {
        "permission": "0o755"
    }
